/*
	 Copyright Ad Lib Inc., 1990

	 This file is part of the Ad Lib Programmer's Manual product and is
	 subject to copyright laws.  As such, you may make copies of this file
	 only for the purpose of having backup copies.  This file may not be
	 redistributed in any form whatsoever.

	 If you find yourself in possession of this file without having received
	 it directly from Ad Lib Inc., then you are in violation of copyright
	 laws, which is a form of theft.
*/

/*
   This program consisits of 2 C files and 3 assembler files.  The C files
   are midimain.c and midiplay.c.  The three assembler files are timer.asm,
   outchip.asm and setfreq.asm.
   Libraries are provided for the assembler files so that you do not
   need an assembler and you do not need to know assembly language.
   The libraries are name xADLIB.LIB where x is S, M, L for Small, Medium
   and Large model code.


   Microsoft C, v5.1 instructions
   ------------------------------
   We suggest compiling with the following options:
	  cl -AL -Zi -J -Ot -Zp -Gs -c


   QuickC 2.0 Instructions
   -----------------------
   The stack checking must be turned off because the timer interrupt is
   used and the interrupt routines use a different stack.
   The files must be compiled with stack checking off (select "Compiler
   Flags" in the "Make" dialog of the "Options" menu) and they must be
   compiled for the release version ("Make" dialog).  QuickC ignores its
   stack check flag if it is not in release version mode.
   In the "Link Flags" dialog (selected in the "Make" dialog), set the
   global flag to ingnore case.  If this is not done, the linker will not
   find the routines in the library.


   Turbo C Instructions
   --------------------
   To compile under Turbo C, go into the file "cflags.h" and de-comment the
   following line so that TURBOC is defined:
	  #define  TURBOC


   The following files must be linked together to make the executable:
	  midimain.obj +
	  midiplay.obj +
	  adlib.obj +
	  timer.obj +
	  outchip.obj +
	  setfreq.obj

*/

#include  "cflags.h"

#include  <stdio.h>
#include  <stdlib.h>
#include  <fcntl.h>
#include  <string.h>
#include  <ctype.h>
#include  <io.h>

/* Keeps track of last change of volume to avoid unnecessary calls to
   change of volume routine. */
extern int current_vol [MAX_VOICES];
extern int volume_flag;


/*-------------------------------------------------------------------------
   Enable or disable the volume, but continue playing the song.
*/
Volume_OnOff (flag)
   int flag;
{
   int n;
   if (flag != volume_flag) {
	  if (!flag)
		 for (n=0; n < MAX_VOICES; n++) {
			SetVoiceVolume (n, 0);
			NoteOff (n);
		 }
	  else
		 for (n=0; n < MAX_VOICES; n++) SetVoiceVolume (n, current_vol [n]);
   }
   volume_flag = flag;
}


/*-------------------------------------------------------------------------
   Read in the midi 1.0 file.
*/
UCHAR *Read_Midi_File (file_name)
   char *file_name;
{
   int file, n, i;
   UCHAR *events, *c;
   long length;

   file = open (file_name, O_RDONLY + O_BINARY);
   if (file < 0) {
	  printf ("Cannot open file %s \n", file_name);
	  return (NULL);
   }

   length = filelength (file);
   if (length > 0xffff) {
	  printf ("File too large.\n");
	  return (NULL);
   }

   /* Allocate buffer for entire file. */
   c = (UCHAR *) malloc (length);
   if (!c) {
	  printf ("Insufficient memory available for reading input file.\n");
	  return (c);
   }
   events = c;

   /* Read the file into the buffer. */
   read (file, c, length);

   close (file);
   return (events);
}

/*-------------------------------------------------------------------------
	Wait until the end of melody  (musRunning == 0).
*/
void  WaitEndMelo()
{
   extern char musRunning;
   char c;
   int n;

   printf ("\nWaiting for end.  Press <ESC> to exit.\n");

   while (musRunning) {
	  #ifndef INT_METHOD
		 Test_Event ();
	  #endif

	  if (kbhit()) {
		  c = getch();
		  if (c == 0x1b) {
			 /* User has pressed ESC.  Shut off all voices.  If they are not
				shut off, instruments that have sustained sounds will
				continue to play. */
			 for (n=0; n < MAX_VOICES; n++) {
				SetVoiceVolume (n, 0);
				NoteOff (n);
			 }
			 Stop_Melo();
			 return;
		  }
		  else {
			 /* User has pressed 'S' to silence the music, but the song
				continues to play.  This is useful in applications where
				you have music playing in the background and you wish to
				allow the user the choice of whether or not he/she wants to
				hear the soundtrack. */
			 c = toupper (c);
			 if (c == 'S') Volume_OnOff (!volume_flag);
		  }
	  }
   }
   printf ("Done\n");
}

/*-------------------------------------------------------------------------
	Midi 1.0 file playback program.
*/
main (argc, argv)
	int argc;
	char * argv[];
{
	extern void Midi_Play (UCHAR *);

	UCHAR *buf;
	int n;

	/* Initalize the low-level sound-driver: */
	if (!SoundColdInit (0x388)) {
		printf ("\nAdlib board not found!");
		exit (1);
	}

	if (argc < 2) {
		fprintf (stderr, "  Format: midiplay midi_file_name \n");
		exit (1);
	}

	/* Read the music file. */
	buf = Read_Midi_File (argv [1]);
	if (!buf) exit (0);

	/* Perform some initialisations ... */
	Midi_Init();
	for (n=0; n < MAX_VOICES; n++) current_vol [n] = 0;

	Midi_Play (buf);

	/* wait until end.... */
	WaitEndMelo();

	/* uninstall the clock driver: */
	Midi_End ();
}


