'
'    Copyright Ad Lib Inc., 1990
'
'    This file is part of the Ad Lib Programmer's Manual product and is
'    subject to copyright laws.  As such, you may make copies of this file
'    only for the purpose of having backup copies.  This file may not be
'    redistributed in any form whatsoever.
'
'    If you find yourself in possession of this file without having received
'    it directly from Ad Lib Inc., then you are in violation of copyright
'    laws, which is a form of theft.
'

' File: MIDIPLAY.BAS

' Program to read and play a Standard MIDI File (1.0).

' March 1990, Dale Glowinski, Ad Lib Inc.

' *** NOTE ***
' To run this in the QB environment, you must have loaded the quick library
' adlib.qlb as follows:
'      qb /ladlib.qlb
' Note that when playing songs while in QB, the tempo will be somewhat
' slower than it should be.  The tempo will correct itself in the
' executable version of this file.

' The driver uses the PC's timer-0 for timing. (see TIMER.ASM )

' Interface instructions:
'    - Call ReadMidiFile to read the file.
'    - Call MidiInit to install the interrupt vector.
'    - Call MidiPlay to play the file.
'    - Call MidiEnd to remove the interrupt vector.
'
'    - You do not need to call MidiInit and MidiEnd every time you
'      play something.  You can call them once at the beginning and
'      end of your program.
'    - The timer-0 is set to generate an interrupt frequency based on the
'      tempo.  The new interrupt routine takes care of chaining to the
'      old interrupt routine at the appropriate moment.
'    - Timer-0 is set to its usual state when the file has finished playing.


' External SUB routines.  The source code for these routines is in the
' file TIMER.ASM.
DECLARE SUB ClkInstall CDECL ()
DECLARE SUB ClkUnInstall CDECL ()
DECLARE SUB StartTimeOut CDECL (BYVAL delay AS INTEGER)
DECLARE SUB SetClkRate CDECL (BYVAL TEMPO AS INTEGER)
DECLARE FUNCTION TestEvent% CDECL ()

'Routines in the module ADLIB.BAS
DECLARE SUB NoteOff (voice AS INTEGER)
DECLARE SUB NoteOn (voice AS INTEGER, pitch AS INTEGER)
DECLARE SUB SetMode (mode AS INTEGER)
DECLARE SUB SetPitchRange (pR AS INTEGER)
DECLARE SUB SetVoicePitch (voice AS INTEGER, pitchBend AS INTEGER)
DECLARE SUB SetVoiceTimbre (voice AS INTEGER, paramArray() AS INTEGER)
DECLARE SUB SetVoiceVolume (voice AS INTEGER, volume AS INTEGER)

'Routines in this module
DECLARE SUB AdLibSpecific (code AS INTEGER, ptr AS INTEGER)
DECLARE SUB DoEvent ()
DECLARE SUB MidiNoteOn (voice%, note%, volume%)
DECLARE SUB MetaEvent ()
DECLARE SUB MidiEvent (event%)
DECLARE SUB MidiInit ()
DECLARE SUB MidiPlay ()
DECLARE SUB MidiEnd ()
DECLARE SUB SetTempo (tickQnote AS INTEGER, usec AS LONG)
DECLARE SUB SetOriginalClock ()
DECLARE SUB StopMelo ()
DECLARE SUB StartMelo (ptr%)
DECLARE SUB SetUpMidiData ()
DECLARE SUB SetUserExit ()
DECLARE SUB SysexEvent (event%)
DECLARE FUNCTION GetWord% (ptr%)
DECLARE FUNCTION Get3Byte& (ptr%)
DECLARE FUNCTION GetLong& (ptr%)
DECLARE FUNCTION GetLength& ()
DECLARE FUNCTION GetNextDelay& ()
DECLARE FUNCTION ReadMidiFile% (fileName AS STRING)
DECLARE FUNCTION SoundColdInit% (portAddress AS INTEGER)

CONST MAXVOICES = 11

' Some midi codes
CONST ENDOFTRACK = &H2F
CONST TEMPO = &H51
CONST SYSEXF0 = &HF0
CONST SYSEXF7 = &HF7
CONST META = &HFF
CONST SEQSPECIFIC = &H7F

DIM SHARED tickQnote AS INTEGER      'ticks per quarter note
DIM SHARED status AS INTEGER         'current running status

'Keeps track of last change of volume to avoid unnecessary calls to the
'change of volume routine.
DIM SHARED currentVol(0 TO MAXVOICES) AS INTEGER

'Flags
DIM SHARED musrunning AS INTEGER     '1 if music is playing, 0 if not
DIM SHARED endOfData AS INTEGER      '1 if end of data, 0 if not
DIM SHARED clockIn AS INTEGER        '1 if clock int S/R installed, 0 if not

'QuickBasic has a limit of 32K on the size of arrays and the size of strings.
'Hence, you cannot read any files longer than 32K.
' $DYNAMIC
DIM SHARED fileBuf(1) AS STRING * 1   'buffer for input file
DIM SHARED fptr AS INTEGER            'pointer into fileBuf


' Main routine for midi playback program begins here.

' Initalize the low-level sound driver:
IF (SoundColdInit(&H388) = 0) THEN
   PRINT "Adlib board not found!"
   END
END IF

' Clear screen and prompt for a file name.
CLS
INPUT "Enter a midi file name: ", fileName$

' Read in the file.  Returns length of file: 0 if no file found.
length% = ReadMidiFile(fileName$)

IF (length% > 0) THEN
   'Set up the ESC key as a user escape sequence
   SetUserExit

   ' Perform some initialisations.  (Installs a clock interrupt routine.)
   MidiInit

   ' Start the song playing.
   MidiPlay

   ' Remove the clock interrupt routine.
   MidiEnd
ELSE
   PRINT "Unable to read file: "; fileName$
END IF

END


'Program jumps to here when user presses ESC.  See also S/R SetUserExit.
UserExit:
   musrunning = 0
   endOfData = 1
   FOR n% = 0 TO MAXVOICES
      CALL SetVoiceVolume(n%, 0)
   NEXT
RETURN

REM $STATIC
'Process an Ad Lib specific meta-event.
'
SUB AdLibSpecific (code AS INTEGER, ptr AS INTEGER)
   IF (code = 1) THEN   'Instrument change code
      DIM n AS INTEGER
      DIM params(0 TO 27) AS INTEGER
      ptr = ptr + 1
      FOR n = 0 TO 27
	 params(n) = ASC(fileBuf(ptr + n))
      NEXT
      CALL SetVoiceTimbre(ASC(fileBuf(ptr - 1)), params())
   ELSEIF (code = 2) THEN    'set melo/perc mode
      SetMode (ASC(fileBuf(ptr)))
   ELSEIF (code = 3) THEN    'set pitch bend range
      SetPitchRange (ASC(fileBuf(ptr)))
   END IF
END SUB

' Process the next MIDI event
SUB DoEvent
   DIM dataByte AS INTEGER
   DIM delay AS INTEGER

   IF (musrunning <> 0) THEN
   DO
      dataByte = ASC(fileBuf(fptr))
    
      'If high bit set, then set status, else this is running status.
      IF (dataByte AND &H80) THEN
	 status = dataByte
	 fptr = fptr + 1
      END IF

      'Process event.
      SELECT CASE status
	 CASE SYSEXF7, SYSEXF0
	    SysexEvent (status)
	 CASE META
	    MetaEvent
	 CASE ELSE
	    MidiEvent (status)
      END SELECT

      'Read next delta time.  A delay time of 0 indicates that the next
      'event occurs at the same time as the just-processed event.
      delay = GetNextDelay
  
   LOOP WHILE (delay = 0 AND endOfData <> 1)
   END IF

   'Tell the timer routine how long it is until the next event occurs.
   'The delay time is in ticks.
   IF (delay <> 0) THEN StartTimeOut (delay)

END SUB

'Get a 3 byte value from data.  The value is stored MSB to LSB.  Note that
'this does not change the global "fptr" variable.
FUNCTION Get3Byte& (ptr%)
   l& = 0
   FOR n% = 0 TO 2
      l& = (l& * 256) + ASC(fileBuf(ptr% + n%))
   NEXT
   Get3Byte& = l&
END FUNCTION

'Reads a variable length value from MIDI file data and advances the global
'data pointer "fptr".
FUNCTION GetLength&
   DIM value AS LONG
   DIM dataByte AS LONG

   value = ASC(fileBuf(fptr))
   IF (value AND &H80) THEN
      value = value AND &H7F
      DO
	 fptr = fptr + 1
	 dataByte = ASC(fileBuf(fptr))
	 value = (value * 128) + (dataByte AND &H7F)
      LOOP UNTIL (NOT (dataByte AND &H80))
   END IF
   fptr = fptr + 1
   GetLength& = value&
END FUNCTION

'Get a 4-byte long value from data.  Value is stored MSB to LSB.
FUNCTION GetLong& (ptr%)
   l& = 0
   FOR n = 0 TO 3
      l& = (l& * 256) + ASC(fileBuf(ptr% + n))
   NEXT
   GetLong& = l&
END FUNCTION

'Returns the time, in ticks, to wait until the next midi event should be
'processed.  Sets various variables if end of data is encountered.
FUNCTION GetNextDelay&
   IF (status <> ENDOFTRACK) THEN
      GetNextDelay& = GetLength                  'get delta time
   ELSE
      'end of data condition
      endOfData = 1
      StopMelo
      GetNextDelay& = 0
   END IF
END FUNCTION

'Get a 2 byte value from data.  Value is stored MSB first.
FUNCTION GetWord% (ptr%)
   l% = ASC(fileBuf(ptr%))
   l% = (l% * 256) + ASC(fileBuf(ptr% + 1))
   GetWord% = l%
END FUNCTION

'Process a meta-event.  All meta-events other than end-of-track, tempo and
'Ad Lib sequencer-specific events are ignored.  The Ad Lib sequencer-
'specific meta-event is explained below.
'   Ad Lib MIDI sequencer specific meta-event format:
'       F0 <length> 00 00 3F <data bytes> F7
'
'   <length> is a variable length quantity.  "00 00 3F" and "F7" are
'   included in the length.
'
'   "00 00 3F" is Ad Lib's MIDI identification code.
'
'   The first two bytes of the data byte section will always be an opcode
'   which identifies the data to follow.  This will be stored as a fixed
'   length quantity (two bytes) and is stored MSB first, LSB second.  The
'   opcode is followed by more data bytes.
'
'   Currently, three opcodes are defined:
'      Opcode    Byte #    Meaning
'      ------    ------    -------
'      00 01               Patch (instrument change)
'                  1       Voice number for instrument
'                2 - 29    28 byte array of parameters
'
'      00 02               Card mode
'                  1       0 = melodic, 1 = percussive
'
'      00 03               Pitch bend range
'                  1       1 - 12 inclusive
'
SUB MetaEvent
   IF (ASC(fileBuf(fptr)) = ENDOFTRACK) THEN
      status = ENDOFTRACK
      fptr = fptr + 1
   ELSEIF (ASC(fileBuf(fptr)) = TEMPO) THEN
      fptr = fptr + 2              'type and length bytes
      l& = Get3Byte(fptr)
      fptr = fptr + 3
      CALL SetTempo(tickQnote, l&)
   ELSEIF (ASC(fileBuf(fptr)) = SEQSPECIFIC) THEN
      DIM length AS LONG
      fptr = fptr + 1
      length = GetLength

      'Test for Ad Lib midi ID (00 00 3F)
      IF (ASC(fileBuf(fptr)) = 0 AND ASC(fileBuf(fptr + 1)) = 0 AND ASC(fileBuf(fptr + 2)) = &H3F) THEN
	 DIM code AS INTEGER
	 code = ASC(fileBuf(fptr + 3))
	 code = (code * 256) OR ASC(fileBuf(fptr + 4))
	 CALL AdLibSpecific(code, fptr + 5)
      END IF
      fptr = fptr + length
   ELSE
      'Read over data.
      fptr = fptr + 1           'type
      fptr = fptr + GetLength   'data
   END IF
END SUB

'Remove the clock interrupt driver.
SUB MidiEnd
   IF (clockIn <> 0) THEN ClkUnInstall
   clockIn = 0
END SUB

'Process a regular MIDI event.  Which routine to call is determined by
'using the 3 LSB's of the high nibble of the passed status byte.  The
'channel (voice) number is in the low nibble.
SUB MidiEvent (event AS INTEGER)
  
   DIM command AS INTEGER
   DIM voice AS INTEGER

   command = (event \ 16) AND 7
   voice = event AND &HF

   IF (voice < MAXVOICES) THEN
      SELECT CASE command
	 CASE 0
	    NoteOff (voice)
	    fptr = fptr + 2
	 CASE 1
	    CALL MidiNoteOn(voice, ASC(fileBuf(fptr)), ASC(fileBuf(fptr + 1)))
	    fptr = fptr + 2
	 CASE 2
	    CALL SetVoiceVolume(voice, ASC(fileBuf(fptr + 1)))
	    currentVol(voice) = ASC(fileBuf(fptr + 1))
	    fptr = fptr + 2
	 CASE 3
	    fptr = fptr + 2
	 CASE 4
	    fptr = fptr + 1
	 CASE 5
	    CALL SetVoiceVolume(voice, ASC(fileBuf(fptr)))
	    currentVol(voice) = ASC(fileBuf(fptr))
	    fptr = fptr + 1
	 CASE 6
	    CALL SetVoicePitch(voice, (ASC(fileBuf(fptr + 1)) * 128) OR ASC(fileBuf(fptr)))
	    fptr = fptr + 2
      END SELECT
   END IF
END SUB

'Install the clock interrupt routine.
SUB MidiInit
   DIM n AS INTEGER
  
   IF (clockIn) THEN EXIT SUB

   ClkInstall
   clockIn = 1
  
   FOR n = 0 TO MAXVOICES
      currentVol(n) = 0
   NEXT
END SUB

'Starts a note playing, changing the volume if necessary.  A note-on with a
'volume of zero if often used to indicate a note-off.
SUB MidiNoteOn (voice AS INTEGER, note AS INTEGER, volume AS INTEGER)
   IF (volume = 0) THEN
      NoteOff (voice)
      currentVol(voice) = 0
   ELSE
      IF (currentVol(voice) <> volume) THEN
	 CALL SetVoiceVolume(voice, volume)
	 currentVol(voice) = volume
      END IF
      CALL NoteOn(voice, note)
   END IF
END SUB

'Main routine for playing a MIDI file.
SUB MidiPlay
  
   IF (clockIn = 0) THEN EXIT SUB
   SetUpMidiData
   PRINT "Press ESC to stop music."
   StartMelo (fptr)
  
   DO
      'TestEvent asks the interrupt driver if it is time to act on an event.
      'DoEvent processes the next event(s) and then tells the interrupt
      'driver how long to wait until the next event occurs.
     
      IF (TestEvent <> 0) THEN DoEvent
  
   LOOP WHILE (endOfData <> 1)

END SUB

'Read in the midi 1.0 file.
'The file is read into the SHARED buffer "fileBuf"
FUNCTION ReadMidiFile% (fileName AS STRING)
   DIM length AS LONG
   DIM n AS INTEGER

   OPEN fileName FOR BINARY AS #1
   length = LOF(1)
   IF (length > 32767) THEN
      'QuickBasic cannot handle arrays and strings larger then 32K
      PRINT "File length exceeds 32767."
      length = 0
   END IF

   IF (length > 0) THEN
      CONST INPUTSIZE = 1024
      DIM localBuf AS STRING * INPUTSIZE
      DIM bufSize AS INTEGER
      DIM index AS INTEGER
      DIM i AS INTEGER
     
      REDIM fileBuf(length) AS STRING * 1

      'Print dots while reading if the file is large.
      IF (length > INPUTSIZE) THEN PRINT "Reading file, please wait";
      n = 0
      WHILE n < length
	 'Reading in blocks of 1024 speeds up the process.  (As opposed
	 'to reading directly into fileBuf byte by byte.)
	 GET #1, , localBuf

	 IF ((n + INPUTSIZE) <= length) THEN
	    bufSize = INPUTSIZE
	 ELSE
	    bufSize = length - n
	 END IF
	   
	 'Copy local buffer to global buffer
	 index = n
	 FOR i = 1 TO bufSize
	    fileBuf(index) = MID$(localBuf, i, 1)
	    index = index + 1
	 NEXT
	   
	 n = n + bufSize
	 IF (length > INPUTSIZE) THEN PRINT ".";
      WEND
      IF (length > INPUTSIZE) THEN PRINT
   END IF
  
   CLOSE #1
   ReadMidiFile = length

END FUNCTION

'Set the clock rate to its original interrupt rate.
SUB SetOriginalClock
   SetClkRate (0)
END SUB

SUB SetTempo (tickQnote AS INTEGER, usec AS LONG)
'       Change the tempo.
'       tickQnote: ticks per quarter note
'       usec:      micro-seconds per quarter note

'       Reload the timer-0 with the proper divider for generating
'       the appropriate interrupt frequency.
'       If tempo is zero, reprogram the counter for 18.2 Hz.
   
    DIM count AS LONG

    IF (tickQnote = 0) THEN
       count = 0
    ELSE
       'Calculate required interrupt rate (ticks per sec)
       usec = usec \ 1000
       count = (1194 * usec) \ tickQnote
    END IF
   
    n% = count
    SetClkRate (n%)
END SUB

'Read header info from midi file.
SUB SetUpMidiData
   'Read file header
   fptr = 4                       'start of header data
   length& = GetLong(fptr)        'header length
   fptr = fptr + 4
  
   format% = GetWord(fptr)        'format (must be 0)
   fptr = fptr + 2
   tracks% = GetWord(fptr)        '# of tracks (must be 1)
   fptr = fptr + 2
   IF (format% <> 0 OR tracks% <> 1) THEN
      PRINT "This is not a MIDI 1.0, format 0 file."
      END
   END IF

   tickQnote = GetWord(fptr)      'ticks per quarter note
   fptr = fptr + 2

   'Read past track header
   fptr = fptr + 8
   abstime& = GetLength           'read first delta time
END SUB

'Set up the escape key so that when it is pressed, the program goes to the
'label "UserExit" in the main module.  The music will then be stopped.
SUB SetUserExit
   KEY 15, CHR$(0) + CHR$(1)
   ON KEY(15) GOSUB UserExit
   KEY(15) ON
END SUB

'Start playing a melody. Set some global pointers and the tempo.  Start
'the music playing by processing the first midi event (call to DoEvent).
SUB StartMelo (ptr%)
   status = ASC(fileBuf(ptr%))
   endOfData = 0
   musrunning = 1

   CALL SetTempo(480, 500000)
   DoEvent
END SUB

'Stop playing the melody. Reset the clock frequency to normal (18.2 Hz).
SUB StopMelo
   musrunning = 0
   SetOriginalClock
END SUB

'Process a midi system exclusive event.
SUB SysexEvent (event AS INTEGER)
   fptr = fptr + GetLength
END SUB

